# Andrej Gelenberg <andrej.gelenberg@udo.edu> (c) 2019

class NXML

  NAME_START = ":A-Z_a-z\u{C0}-\u{D6}\u{D8}-\u{F6}\u{F8}-\u{2FF}\u{370}-\u{37D}\u{37F}-\u{1FFF}\u{200C}-\u{200D}\u{2070}-\u{218F}\u{2C00}-\u{2FEF}\u{3001}-\u{D7FF}\u{F900}-\u{FDCF}\u{FDF0}-\u{FFFD}\u{10000}-\u{EFFFF}"
  NAME_CHAR = "\\-.0-9\u{B7}\u{0300}-\u{036F}\u{203F}-\u{2040}"

  NAME_RE = Regexp.new "^[#{NAME_START}][#{NAME_START}#{NAME_CHAR}]*$"

  ATTR_ESCAPE_HASH = {
    '&' => '&amp;',
    '"' => '&quot;',
    '<' => '&lt;'
  }
  ATTR_ESCAPE = Regexp.new "[#{ATTR_ESCAPE_HASH.keys.join}]"

  TEXT_ESCAPE_HASH = {
    '&' => '&amp;',
    '<' => '&gt;',
    '>' => '&lt;'
  }
  TEXT_ESCAPE = Regexp.new "[#{TEXT_ESCAPE_HASH.keys.join}]"

  def self.check_name(name)
    if not name.match? NAME_RE
      raise ArgumentError.new "invalid XML name \"#{name}\""
    end
  end

  def self.escape_attr_value(value)
    value.gsub ATTR_ESCAPE, ATTR_ESCAPE_HASH
  end

  def self.escape_attr_raw(value)
    value.gsub /["]/, '&quot;'
  end

  def self.escape_text(value)
    value.gsub TEXT_ESCAPE, TEXT_ESCAPE_HASH
  end

  def self.escape_text_raw(value)
    value.gsub /[<]/, '&lt;'
  end

  class Text
    def initialize(text, raw = false)
      if raw
        @text = NXML::escape_text_raw text
      else
        @text = NXML::escape_text text
      end
    end

    def to_str
      @text
    end
  end

  class Element
    def initialize(name, &block)
      NXML::check_name name

      @name       = name
      @attributes = {}
      @children   = []

      instance_eval &block if block_given?
    end

    def method_missing(name, value = nil, *attrs)
      if name[-1] == '='
        name = name[0..-2]
        if value.nil?
          @attributes[name]
        else
          attribute name, value
        end
      else
        super name, value, *attrs
      end
    end

    def tag(name, &block)
      r = Element.new(name, &block)
      @children << r
      r
    end

    def text(text)
      r = Text.new(text)
      @children << r
      r
    end

    def cdata(text)
      r = CDATA.new(text)
      @children << r
      r
    end

    def attribute(name, value)
      NXML::check_name name
      r = NXML::escape_attr_value value
      @attributes[name] = r
      r
    end

    def attribute_raw(name, value)
      XML::check_name name

      r = NXML::escape_attr_raw value
      @attributes[name] = r
      r
    end

    def a(name, value)
      attribute name, value
    end

    def to_str
      ret = "<#{@name}"

      if not @attributes.empty?
        for k,v in @attributes
          ret += " #{k}=\"#{v}\""
        end
      end

      if @children.empty?
        ret += "/>"
      else
        ret += ">#{@children.join}</#{@name}>"
      end

      ret
    end

  end

  class CDATA
    def initialize(data)
      if not data.index(']]>').nil?
        raise ArgumentError.new "CDATA should not contain CDATA closing sequence ']]>'"
      end

      @data = data
    end

    def to_str
      "<![CDATA[#{@data}]]>"
    end
  end

  attr_accessor :version, :encoding, :standalone, :root

  def initialize(&block)
    version = "1.0"
    @encoding = Encoding::UTF_8

    instance_eval &block
  end

  def root(tagname, &block)
    @root = Element.new tagname, &block
  end

  def to_str
    ret = "<?xml"

    if @root.nil?
    end

    if not @version.nil?
      ret += " version=\"#{@version}\""
    end

    if not @encoding.nil?
      ret += " encoding=\"#{@encoding.name}\""
    end

    if not @standalone.nil?
      ret += " standalone=\"#{@standalone ? "yes" : "no"}\""
    end

    ret += "?>\n"

    ret += @root

    ret.encode @encoding
  end
end
